@@ -19,8 +19,8 @@
 ;; URL:      https://github.com/meedstrom/org-mem
 ;; Created:  2025-03-15
 ;; Keywords: text
-;; Package-Version: 20251015.1247
-;; Package-Revision: 7c05f83accaa
+;; Package-Version: 20251021.1607
+;; Package-Revision: c3a0ba6a3aa4
 ;; Package-Requires: ((emacs "29.1") (el-job "2.4.8") (llama "0.5.0"))
 
 ;;; Commentary:
@@ -205,7 +205,6 @@ These are hooks called many times:
 - `org-mem-record-link-functions'
 - `org-mem-forget-file-functions'
 - `org-mem-forget-entry-functions'
-- `org-mem-forget-link-functions'
 
 This lets a function on these hooks sidestep the performance overhead of
 `with-temp-buffer' or `with-work-buffer', in favor of using the
@@ -251,6 +250,10 @@ in `org-link-plain-re').  In practice, it is often an org-id like
 
 NOTE! A future version may omit the sigil @ in citekeys.")
 
+(defvar org-mem--target<>old-links (make-hash-table :test 'equal)
+  "Previous state of `org-mem--target<>links'.
+For downstream use.")
+
 (defvar org-mem--internal-entry-id<>links (make-hash-table :test 'eq)
   "1:N table mapping internal entry ID to list of `org-mem-link' records.
 The list represents all links found in that entry,
@@ -339,8 +342,7 @@ perfectly with `org-id-locations'."
 (defun org-mem-all-files-expanded ()
   "All Org files, with tilde expansion applied."
   (with-memoization (org-mem--table 0 'org-mem-all-files-expanded)
-    (with-temp-buffer ;; No buffer-env
-      (mapcar #'expand-file-name (org-mem-all-files)))))
+    (mapcar #'directory-abbrev-apply (org-mem-all-file-truenames))))
 
 (defun org-mem-all-files ()
   "All Org files."
@@ -624,7 +626,7 @@ use `org-mem-entry-olpath-with-file-title-or-basename' instead."
           (pop olp))
         olp))))
 
-;; Better than above: single-argument func.
+;; Better than above; unary function.
 (defun org-mem-entry-olpath-with-file-title-or-basename (entry)
   "Outline path to ENTRY, including file #+title.
 Use file basename if there is no #+title."
@@ -1197,9 +1199,6 @@ hence the name.  Contrast `org-mem-post-full-scan-functions'.")
 (defvar org-mem-forget-entry-functions nil
   "Hook passed one forgotten `org-mem-entry' object.")
 
-(defvar org-mem-forget-link-functions nil
-  "Hook passed one forgotten `org-mem-link' object.")
-
 (defvar org-mem--problems nil)
 (defvar org-mem--title-collisions nil)
 (defvar org-mem--id-collisions nil)
@@ -1293,7 +1292,6 @@ overrides a default message printed when `org-mem-do-cache-text' is t."
   (clrhash org-mem--truename<>metadata)
   (clrhash org-mem--truename<>entries)
   (clrhash org-mem--internal-entry-id<>links)
-  (clrhash org-mem--target<>links)
   (setq org-mem--title-collisions nil)
   (seq-let (bad-paths file-data entries links problems) parse-results
     (when bad-paths
@@ -1303,8 +1301,14 @@ overrides a default message printed when `org-mem-do-cache-text' is t."
       (dolist (fdata file-data)
         (puthash (car fdata) fdata org-mem--truename<>metadata)
         (run-hook-with-args 'org-mem-record-file-functions fdata))
-      (mapc #'org-mem--record-entry entries)
-      (mapc #'org-mem--record-link links))
+      (dolist (entry entries)
+        (org-mem--record-entry entry)
+        (run-hook-with-args 'org-mem-record-entry-functions entry))
+      (dolist (link links)
+        (push link (gethash (org-mem-link--internal-entry-id link)
+                            org-mem--internal-entry-id<>links))
+        (run-hook-with-args 'org-mem-record-link-functions link)))
+    (org-mem--rebuild-specially-indexed-tables)
 
     (setq org-mem--time-elapsed
           (float-time (time-since org-mem--time-at-begin-full-scan)))
@@ -1345,13 +1349,6 @@ overrides a default message printed when `org-mem-do-cache-text' is t."
       (setq org-mem--problems (append problems org-mem--problems))
       (message "Scan had problems, see M-x org-mem-list-problems"))))
 
-(defun org-mem--record-link (link)
-  "Add info related to LINK to various tables."
-  (push link (gethash (org-mem-link-target link) org-mem--target<>links))
-  (push link (gethash (org-mem-link--internal-entry-id link)
-                      org-mem--internal-entry-id<>links))
-  (run-hook-with-args 'org-mem-record-link-functions link))
-
 (defun org-mem--record-entry (entry)
   "Add info related to ENTRY to various tables."
   (let ((id (org-mem-entry-id entry))
@@ -1362,6 +1359,7 @@ overrides a default message printed when `org-mem-do-cache-text' is t."
     (push entry (gethash truename org-mem--truename<>entries))
     (when id
       (org-mem--maybe-snitch-to-org-id entry)
+      (puthash id entry org-mem--id<>entry)
       (when title
         (let ((other-id (gethash title org-mem--title<>id)))
           (when (and other-id (not (string= id other-id)))
@@ -1370,9 +1368,21 @@ overrides a default message printed when `org-mem-do-cache-text' is t."
         ;; REVIEW: Should we fallback on file name, i.e. use
         ;; `org-mem-entry-title' instead of `org-mem-entry-title-maybe'?
         ;; It would affect function `org-mem-id-by-title'.
-        (puthash title id org-mem--title<>id))
-      (puthash id entry org-mem--id<>entry))
-    (run-hook-with-args 'org-mem-record-entry-functions entry)))
+        (puthash title id org-mem--title<>id)))))
+
+;; A "specially indexed" table is one that cannot be trivially updated when
+;; one file is re-scanned, since a given key in it may list values from
+;; multiple files.  So rather than try to find the particular values to
+;; remove, it's easiest to wipe and re-build.
+;; Note these tables are basically re-expressions of info that already exists.
+(defun org-mem--rebuild-specially-indexed-tables ()
+  "Rebuild table `org-mem--target<>links'."
+  (setq org-mem--target<>old-links (copy-hash-table org-mem--target<>links))
+  (clrhash org-mem--target<>links)
+  (cl-loop
+   for links being each hash-value of org-mem--internal-entry-id<>links
+   do (dolist (link links)
+        (push link (gethash (org-mem-link-target link) org-mem--target<>links)))))
 
 (defun org-mem--maybe-snitch-to-org-id (entry)
   "Add applicable ENTRY data to `org-id-locations'.
@@ -1786,7 +1796,7 @@ org-id-locations:
       (org-mem--scan-full))))
 
 
-(defvar org-mem--bump-int 17 "Not a version number, but bumped sometimes.")
+(defvar org-mem--bump-int 18 "Not a version number, but bumped sometimes.")
 (defmacro org-mem--def-whiny-alias (old new when removed-by)
   "Define function OLD as effectively an alias for NEW.
 Also, calling OLD will emit a deprecation warning the first time.
@@ -1821,6 +1831,7 @@ may be removed from the package."
 
 (define-obsolete-function-alias 'org-mem-entry-olpath-with-file-title-with-self #'org-mem-entry-olpath-with-self-with-file-title  "0.22.0 (2025-10-01)")
 (define-obsolete-function-alias 'org-mem-olpath-with-file-title-with-self       #'org-mem-olpath-with-self-with-file-title        "0.22.0 (2025-10-01)")
+(defvar org-mem-forget-link-functions :obsolete) ;; 2025-10-15 but unused for long already
 
 (provide 'org-mem)
 
